/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann           *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "SearchController.h"

/**
 * \brief Sets the TableModel and the View and initializes the attributes.
 *
 * \param parent Pointer to the parent QObject.
 * \param model Pointer to the instance of TableModel that contains the data.
 * \param view Pointer to the instance of UCQTableView that will be used to
 *             display hits.
 */
SearchController::SearchController ( QObject* parent, TableModel* model,
                                     UCQTableView* view ) :
    Controller ( parent )
{
	setName("SearchController");

    model_ = model;
    view_  = view;

    findindex_ = -1;

    regexp_          = false;
    comment_         = false;
    collation_       = false;
    casesensitivity_ = false;
    codepoint_       = true;

    qRegisterMetaType<QModelIndexList> ( "QModelIndexList" );
    qRegisterMetaType<QList<Letter*> > ( "QList<Letter*>" );

    connect ( this, SIGNAL ( selectLetter      ( QModelIndex, bool ) ),
              view, SLOT   ( setLetterSelected ( QModelIndex, bool ) ),
              Qt::QueuedConnection
            );

    connect ( this, SIGNAL ( scrollTo( QModelIndex ) ),
              view, SLOT   ( scrollTo( QModelIndex ) ),
              Qt::QueuedConnection
            );

    connect ( this, SIGNAL ( clearSelection() ),
              view, SLOT   ( clearSelection() ),
              Qt::QueuedConnection
            );

    connect ( this, SIGNAL ( selectLetters      ( QModelIndexList, bool ) ),
              view, SLOT   ( setLettersSelected ( QModelIndexList, bool ) ),
              Qt::QueuedConnection
            );

    connect ( this,  SIGNAL ( addLetters( QList<Letter*> ) ),
              model, SLOT   ( addList   ( QList<Letter*> ) ),
              Qt::BlockingQueuedConnection
            );
}

/**
 * \brief Adds all hits in the tableView.
 *
 * The hits are appended to the visible table.
 */
void SearchController::addItems ()
{
    emit addLetters ( findlist_ );
}

/**
 * \brief Clears the search.
 */
void SearchController::clear ()
{
    findindex_  = -1;
    searchtext_ = "";

    regexp_          = false;
    comment_         = false;
    collation_       = false;
    casesensitivity_ = false;
    codepoint_       = true;

    findlist_.clear ();
    lvls_.clear ();
}

/**
 * \brief Getter for the case sensitive attribute.
 *
 * \return If true, search is set case sensitive.
 */
bool SearchController::getCasesensitivity () const
{
    return casesensitivity_;
}

/**
 * \brief Getter for the codepoint attribute.
 *
 * \return If true, codepoints are included in search.
 */
bool SearchController::getCodepoint () const
{
    return codepoint_;
}

/**
 * \brief Getter for the collation attribute.
 *
 * \return If true, collations are included in search.
 */
bool SearchController::getCollation () const
{
    return collation_;
}

/**
 * \brief Getter for the comment attribute.
 *
 * \return If true, comments are included in search.
 */
bool SearchController::getComment () const
{
    return comment_;
}

/**
 * \brief Getter for the collation levels attribute.
 *
 * \return List with level flags.
 */
QList<bool> SearchController::getLvls () const
{
    return lvls_;
}

/**
 * \brief Getter for the regexp attribute.
 *
 * \return If true, the searchtext is regarded as regex.
 */
bool SearchController::getRegexp () const
{
    return regexp_;
}

/**
 * \brief Getter for the searchtext.
 *
 * \return Current Searchtext.
 */
QString SearchController::getSearchtext () const
{
    return searchtext_;
}

/**
 * \brief Checks if there are more hits.
 *
 * \return If true, there is at least one more hit.
 */
bool SearchController::hasNext ()
{
    return findindex_ + 1 < findlist_.count ();
}

/**
 * \brief Checks if there are previous hits.
 *
 * \return If true, there is at least one previous hit.
 */
bool SearchController::hasPrevious ()
{
    return findindex_ > 0;
}

/**
 * \brief Checks if search was unsuccessful.
 *
 * \return If true, nothing was found.
 */
bool SearchController::isEmpty ()
{
    return findlist_.count () == 0;
}

/**
 * \brief Selects the next hit in the tableView.
 *
 * If there are no next hits nothing happens.
 */
void SearchController::next ()
{
    if ( !hasNext () )
    {
        return;
    }
    findindex_++;

    emit clearSelection ();

    QModelIndex i = model_->index ( findlist_.at ( findindex_ ) );
    emit selectLetter ( i, true );
    emit scrollTo ( i );

}

/**
 * \brief Get count of found items.
 *
 * \return Number of hits.
 */
int SearchController::numberOfHits ()
{
    return findlist_.count ();
}

/**
 * \brief Selects the previous hit in the tableView.
 *
 * If there are no previous hits nothing happens.
 */
void SearchController::previous ()
{
    if ( !hasPrevious () )
    {
        return;
    }

    findindex_--;

    emit clearSelection ();

    QModelIndex i = model_->index ( findlist_.at ( findindex_ ) );
    emit selectLetter ( i, true );
    emit scrollTo ( i );

}
/**
 * \brief This method searches in the currently visible table provided by
 * the model.
 *
 * The parameters determine which attributes of the letters are taken into
 * account for the search and if the search will be case sensitive or allow
 * regular expressions.
 * To search the whole dataset of the tablemodel see searchTable().
 *
 * \param searchtext String or QRegExp to search.
 * \param codepoint If true, search in codepoints.
 * \param collation If true, search in collations.
 * \param comment If true, search in comments.
 * \param casesensitivity If true, the search is case sensitive.
 * \param regexp If true, the serachtext is regarded as QRegExp.
 * \param lvls In a search with collations == true this parameter determines
 *             which collationlevels (0-3) are taken into account.
 *
 * \return If true, the search was successful and there are hits.
 * \see next()
 * \see previous()
 * \see selectAll()
 * \see searchForRegExp()
 * \see searchForString()
 */
bool SearchController::search ( QString searchtext, bool codepoint,
                                bool collation, bool comment, bool linestring,
                                bool casesensitivity, bool regexp,
                                QList<bool> & lvls)
{
    clear ();

    // set attributes for searchString() or searchForRegExp()
    searchtext_      = searchtext;
    comment_         = comment;
    collation_       = collation;
    casesensitivity_ = casesensitivity;
    regexp_          = regexp;
    codepoint_       = codepoint;
    lvls_            = lvls;
    linestring_ 	 = linestring;
    letters_ = model_->getVisibleItems ();

    if ( regexp_ )
    {
        findlist_ = searchForRegExp ();
    } else
    {
        findlist_ = searchForString ();
    }

    return findlist_.count () > 0;
}

/**
 * \brief Searches for regex matches on the searchstring.
 *
 * The search is delegated to this method from search() and searchTable(),
 * if their regexp arguments are true.
 *
 * \return List of hits.
 */
QList<Letter*> SearchController::searchForRegExp ()
{
    QRegExp rx ( searchtext_ );
    rx.setCaseSensitivity ( Qt::CaseInsensitive );
    QList<Letter*> findlist;
    Letter* letter;
    for ( int i = 0; i < letters_.size (); ++i )
    {
        letter = 0;
        letter = letters_.at ( i );
        if ( letter == 0 )
            continue;


        if ( codepoint_ )
        {
            QString codepointAsHex = letter->getCodepointAsHex ();

            if ( rx.indexIn ( codepointAsHex.toUpper (), 0 ) != -1 )
            {
                findlist << letter;
                continue;
            }
        }
        if ( collation_ )
        {
            QList<Collation> collations = letter->getCollations ();

            for ( int k = 0; k < collations.count (); k++ )
            {
                Collation col = collations.at ( k );
                QList<uint> weights = col.getWeights ();
                QString tmp;
                for ( int l = 0; l < 4; l++ )
                {
                    if ( lvls_.at ( l ) == true )
                    {

                        tmp = QString::number ( weights.at ( l ), 16 );

                        while ( tmp.length () < 4 )
                        {
                            tmp.prepend ( '0' );
                        }

                        if ( rx.indexIn ( tmp.toUpper (), 0 ) != -1 )
                        {
                            findlist << letter;
                            break;
                        }

                    }
                }
                if ( !lvls_.contains ( true ) )
                {
                    QString collationString = collations.at ( k ).toString ();
                    if ( rx.indexIn ( collationString, 0 ) != -1 )
                    {
                        findlist << letter;
                        continue;
                    }
                }
                if ( findlist.contains ( letter ) )
                {
                    continue;
                }
            }
        }
        if ( comment_ )
        {
            QString codepointComment = letter->getComment ();

            if ( casesensitivity_ )
            {
                rx.setCaseSensitivity ( Qt::CaseSensitive );
            } else
            {
                codepointComment = codepointComment.toUpper ();
            }

            if ( rx.indexIn ( codepointComment, 0 ) != -1 )
            {
                findlist << letter;
                continue;
            }
            rx.setCaseSensitivity ( Qt::CaseInsensitive );
        }
        if(linestring_)
        {
        	 QString linestring = letter->getLinestring();
        	  if ( casesensitivity_ )
        	  {
        		  rx.setCaseSensitivity ( Qt::CaseSensitive );
        	  }
        	  else
        	  {
        		  linestring = linestring.toUpper ();
        	  }
        	  if ( rx.indexIn ( linestring, 0 ) != -1 )
        	  {
        		  findlist << letter;
        		  continue;
        	  }
        }
    }
    return findlist;
}

/**
 * \brief Searches for exact matches on the searchstring.
 *
 * The search is delegated to this method by search() and searchTable(),
 * if their regexp arguments are false.
 *
 * \return List of hits.
 */
QList<Letter*> SearchController::searchForString ()
{
    Letter* letter;
    QList<Letter*> findlist;
    for ( int i = 0; i < letters_.size (); ++i )
    {
        letter = 0;
        letter = letters_.at ( i );
        if ( letter == 0 )
            continue;
        if ( codepoint_ )
        {
            QString codepointAsHex = letter->getCodepointAsHex ();

            if ( searchtext_.toUpper () == codepointAsHex.toUpper () )
            {
                findlist << letter;
                continue;
            }
        }

        if ( collation_ )
        {
            QList<Collation> collations = letter->getCollations ();

            for ( int k = 0; k < collations.count (); k++ )
            {
                Collation col = collations.at ( k );
                QList<uint> weights = col.getWeights ();
                QString tmp;
                for ( int l = 0; l < 4; l++ )
                {
                    if ( lvls_.at ( l ) == true )
                    {
                        tmp = QString::number ( weights.at ( l ), 16 );

                        while ( tmp.length () < 4 )
                        {
                            tmp.prepend ( '0' );
                        }

                        if ( searchtext_.toUpper () == tmp.toUpper () )
                        {
                            findlist << letter;
                            break;
                        }
                    }
                }
                if ( !lvls_.contains ( true ) )
                {
                    QString collationString = collations.at ( k ).toString ();
                    if ( searchtext_ == collationString.toUpper () )
                    {
                        findlist << letter;
                        continue;
                    }
                }
                if ( findlist.contains ( letter ) )
                {
                    continue;
                }
            }
        }
        if ( comment_ )
        {
            QString codepointComment = letter->getComment ();

            if ( !casesensitivity_ )
            {
                codepointComment = codepointComment.toUpper ();
                searchtext_ = searchtext_.toUpper ();
            }

            if ( searchtext_ == codepointComment )
            {
                findlist << letter;
                continue;
            }
        }
        if (linestring_)
        {
			QString linestring = letter->getComment();

			if (!casesensitivity_) {
				linestring = linestring.toUpper();
				searchtext_ = searchtext_.toUpper();
			}

			if (searchtext_ == linestring) {
				findlist << letter;
				continue;
			}
		}
    }
    return findlist;
}

/**
 * \brief This method searches in the whole dataset of the model.
 *
 * It accepts the same arguments as search(), but the hits are not available for
 * next(), previous() or selectAll(). They are instead appended to the visible
 * table of the TableModel.
 *
 * \param searchtext String or QRegExp to search.
 * \param codepoint If true, search in codepoints.
 * \param collation If true, search in collations.
 * \param comment If true, search in comments.
 * \param casesensitivity If true, the search is case sensitive.
 * \param regexp If true, the serachtext is regarded as QRegExp.
 * \param lvls In a search with collations == true this parameter determines
 *             which collationlevels (0-3) are taken into account.
 *
 * \return If true, the search was successful and there are hits.
 * \see searchForRegExp()
 * \see searchForString()
 */
void SearchController::searchTable ( QString searchtext, bool codepoint,
                                     bool collation, bool comment, bool linestring,
                                     bool casesensitivity, bool regexp, QList<
                                             bool> & lvls )
{

    clear ();
    searchtext_      = searchtext;
    comment_         = comment;
    collation_       = collation;
    linestring_		 = linestring;
    casesensitivity_ = casesensitivity;
    regexp_          = regexp;
    codepoint_       = codepoint;
    lvls_            = lvls;

    letters_ = model_->getItems ();

    if ( regexp_ )
    {
        model_->addList ( searchForRegExp () );
    } else
    {
        model_->addList ( searchForString () );
    }

    clear ();
}

/**
 * \brief Selects all hits in the tableView.
 *
 * If there are no hits nothing happens.
 */
void SearchController::selectAll ()
{
    if ( findlist_.isEmpty () )
        return;emit
    clearSelection ();

    QModelIndexList selectionList;
    for ( int i = 0; i < findlist_.count (); i++ )
    {
        QApplication::processEvents ();
        selectionList.append ( model_->index ( findlist_.at ( i ) ) );
    }
    emit selectLetters ( selectionList, true );
}

/**
 * \brief Setter for the case sensitive attribute.
 *
 * \param casesensitivity If search shall be case sensitiv.
 */
void SearchController::setCasesensitivity ( bool casesensitivity )
{
    casesensitivity_ = casesensitivity;
}

/**
 * \brief Setter for the codepoint attribute.
 *
 * \param codepoint If search shall include codepoints.
 */
void SearchController::setCodepoint ( bool codepoint )
{
    codepoint_ = codepoint;
}

/**
 * \brief Setter for the collation attribute.
 *
 * \param collation If search shall include collations.
 */
void SearchController::setCollation ( bool collation )
{
    collation_ = collation;
}

/**
 * \brief Setter for the comment attribute.
 *
 * \param comment If search shall include comments.
 */
void SearchController::setComment ( bool comment )
{
    comment_ = comment;
}

/**
 * \brief Setter for the collation levels attribute.
 *
 * \param lvls Flags for levels to include in the search.
 */
void SearchController::setLvls ( QList<bool> lvls )
{
    lvls_ = lvls;
}

/**
 * \brief Setter for the regexp attribute.
 *
 * \param regexp If the searchtext is a regex.
 */
void SearchController::setRegexp ( bool regexp )
{
    regexp_ = regexp;
}

/**
 * \brief Setter for the searchtext.
 *
 * \param searchtext Searchtext.
 */
void SearchController::setSearchtext ( QString searchtext )
{
    searchtext_ = searchtext;
}

/**
 * \brief Select all blocks.
 */
void SearchController::setAllBlocks(){
	model_->showAll();
}
