/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann		   *
 *      																   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "SortController.h"
#include <QDebug>
#include <QListIterator>
/**
 * \brief Constructor for the SortController.
 *
 * \param parent Pointer to the parent QObject.
 * \param model Pointer to the instance of TableModel that contains the data.
 */
SortController::SortController ( QObject* parent, TableModel * model ) :
    Controller ( parent )
{
    setName ( "SortController" );

    model_ = model;
    //    defaultList = model_->getVisibleItems ();

    this->sa_ = ASCENDING;

    qRegisterMetaType<QList<Letter*> > ( "QList<Letter*>" );

    connect ( this, SIGNAL ( toggleTableSignal (QList<Letter*>) ), this,
              SLOT ( reverseTable (QList<Letter*>) ), Qt::QueuedConnection );

    connect ( this, SIGNAL ( clear (bool) ), model_, SLOT ( clearVisible (bool) ),
              Qt::QueuedConnection );

    connect ( this, SIGNAL ( addLetters ( QList<Letter*> ) ), model,
              SLOT ( addList ( QList<Letter*> ) ), Qt::QueuedConnection );

}

/**
 * \brief Sorts the visible letters by codepoints.
 *
 * \param sm The mode of sorting to use.
 *
 * \see codepointLessThan()
 * \see commentLessThan()
 * \see collationLessThan()
 * \see TableModel::getVisibleItems()
 */
void SortController::sort ( SortMode sm )
{
    QApplication::setOverrideCursor ( Qt::WaitCursor );
    QList<Letter*> visible = model_->getVisibleItems ();
    switch ( sm )
    {
        case CODEPOINT_SORT:
            qSort ( visible.begin (), visible.end (),
                    SortController::codepointLessThan );
            break;
        case COLLATION_SORT:
            qSort ( visible.begin (), visible.end (),
                    SortController::collationLessThan );
            break;
        case COMMENT_SORT:
            qSort ( visible.begin (), visible.end (),
                    SortController::commentLessThan );
            break;
    }
    if ( this->sa_ == DESCENDING )
    {
        emit toggleTableSignal ( visible );
    } else
    {
        emit clear (false);
        emit addLetters ( visible );
    }
    QApplication::restoreOverrideCursor ();
}

/**
 * \brief Changes the order of the table elements.
 *
 * \param sa The order of sorting (ascending or descending).
 */
void SortController::sortOrder ( SortOrder sa )
{
    QList<Letter*> visible = model_->getVisibleItems ();
    if ( sa != this->sa_ )
    {
        emit toggleTableSignal ( visible );
        this->sa_ = sa;
    }
}
/**
 * \brief Reverses the sorting of given QList and puts the contents in the
 *        table.
 *
 * \param letters The list to reverse and display.
 */
void SortController::reverseTable ( QList<Letter*> letters )
{
    QList<Letter*> tmp;

    QListIterator<Letter*> it ( letters );
    for ( it.toBack (); it.hasPrevious (); )
    {
        tmp << it.previous ();
    }
    letters = QList<Letter*> ( tmp );
    emit clear (false);
    emit addLetters ( letters );
}

/**
 * \brief Checks if the collation of the first letter is smaller than the
 *        collation of the second letter. Letters no collation,
 *        are inserted at the end of the table, if the order ascending is
 *        given. Otherwise the letters are inserted at the beginning of the table.
 *
 * \param l1 First Letter.
 * \param l2 Second Letter.
 * \return True if l1 codepoint is smaller than l2 codepoint.
 */
bool SortController::collationLessThan ( const Letter* l1, const Letter* l2 )
{
    QList<Collation> list1 = l1->getCollations ();
    QList<Collation> list2 = l2->getCollations ();
    if ( list1.isEmpty () && list2.isEmpty () )
    {
        return codepointLessThan ( l1, l2 );
    } else if ( list1.isEmpty () )
    {
        return false;
    } else if ( list2.isEmpty () )
    {
        return true;
    } else
    {
        for ( int i = 0; i < list2.count (); i++ )
        {
            if ( list1.count () == i )
            {
                return true;
            }
            Collation c1 = list1.at ( i );
            Collation c2 = list2.at ( i );

            if ( !c1.isEmpty () && !c2.isEmpty () )
            {
                if ( c1.getWeights ().first () == c2.getWeights ().first () )
                {
                    continue;
                }
                return c1.getWeights ().first () < c2.getWeights ().first ()
                                                                             ? true
                                                                             : false;
            }
            //            for ( int j = 0; j < c2.getWeights ().count (); j++ )
            //            {
            //                if( c1.getWeights().size() == j ){
            //                    return true;
            //                }
            //                if ( c1.getWeights ().at ( j ) == c2.getWeights ().at ( j ) )
            //                {
            //                    continue;
            //                } else if ( c1.getWeights ().at ( j )
            //                        < c2.getWeights ().at ( j ) )
            //                {
            //                    return true;
            //                } else
            //                {
            //                    return false;
            //                }
            //            }
        }
        return false;
    }
}

/**
 * \brief Checks if the codepoint of the first letter is smaller than the
 *        codepoint of the second letter.
 *
 * \param l1 First Letter.
 * \param l2 Second Letter.
 * \return True if l1 codepoint is smaller than l2 codepoint.
 */
bool SortController::codepointLessThan ( const Letter * l1, const Letter* l2 )
{
    int size = 0;
    int l1size = l1->getCodepoint ().size ();
    int l2size = l2->getCodepoint ().size ();

    size = ( l1size < l2size ) ? l1size : l2size;

    int cntequal = 0;
    for ( int i = 0; i < size; i++ )
    {
        if ( l1->getCodepoint ().at ( i ) < l2->getCodepoint ().at ( i ) )
            return true;
        if ( l1->getCodepoint ().at ( i ) == l2->getCodepoint ().at ( i ) )
            cntequal++;
    }

    if ( cntequal == l1size )
    {
        return true;
    }

    return false;
}

/**
 * \brief Checks if the comment of the first letter is smaller than the
 *        comment of the second letter.
 *
 * \param l1 First Letter.
 * \param l2 Second Letter.
 * \return True if l1 codepoint is smaller than l2 codepoint.
 */
bool SortController::commentLessThan ( const Letter * l1, const Letter* l2 )
{
    QString com1 = l1->getComment ();
    QString com2 = l2->getComment ();

    return com1 < com2 ? true : false;
}

///**
// * \brief Restore the standard order.
// */
//QList<Letter *> SortController::restoreDefault ()
//{
//    //    defaultList = QList<Letter*> ( model_->getVisibleItems () );
//
//    QList<Letter*> defaultList = model_->getItems ();
//    QList<Letter*> visList = model_->getVisibleItems ();
//    for ( int i = 0; i < defaultList.size (); ++i )
//    {
//        if ( !visList.contains ( defaultList.at ( i ) ) )
//        {
//            defaultList.removeAt ( i );
//            --i;
//        }
//    }
//    return defaultList;
//}
