/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann           *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef TABLECONTROLLER_H_
#define TABLECONTROLLER_H_

#include <QObject>
#include "data/TableModel.h"
#include "controller/Controller.h"

/**
 * \class TableController
 * \brief Controller for the use of the TableModel from parallel threads,
 *        i.e. Lua functions.
 *
 * All methods of this class emit signals to the TableModel, to make it
 * possible to change the tableModel from another thread.
 *
 * \author Martin Burgemeister <Martin.Burgemeister@gmx.de>,
 *         Mathis Lindemann <MathisLindemann@gmx.net>
 * \version 1.0
 * \date 16 August 2010
 * \see TableModel
 */
class TableController: public Controller
{

Q_OBJECT

public:
    TableController ( QObject* parent, TableModel* model );

private:
    /**
     * \brief Instance of TableModel that will be connected to the signals.
     */
    TableModel * model_;

public slots:
    void clearTable ();

    void selectBlocks ( QStringList, bool );

    void addRange  ( uint, uint );
    void hideRange ( uint, uint );

    void addWeightRange ( QStringList, QStringList, TableModel::CollationModes,
                          QList<int> );

signals:
    /**
     * \brief Signal to clear the visible items of the model.
     *
     * \see clearTable()
     * \see TableModel::clearVisible()
     */
    void clear ();
    /**
     * \brief Signal to select a list of blocks.
     *
     * \param blocks Names of the blocks to select.
     * \param select If true the blocks are selected.
     *
     * \see selectBlocks()
     * \see UCQBlockList::selectBlocks()
     */
    void selectBlocksSignal ( QStringList blocks, bool select);
    /**
     * \brief Signal to add a range of letters.
     *
     * \param start First codepoint of range to add.
     * \param end Last codepoint of range to add.
     *
     * \see addRange()
     * \see TableModel::addRange()
     */
    void addRangeSignal ( uint start, uint end );
    /**
     * \brief Signal to hide a range of letters.
     *
     * \param start First codepoint of range to hide.
     * \param end Last codepoint of range to hide.
     *
     * \see hideRange()
     * \see TableModel::hideRange()
     */
    void hideRangeSignal ( uint start, uint end );

    /**
     * \brief Signal to add a range of weights.
     *
     * \param begin First/lowest collation.
     * \param end Last/highest collation.
     * \param flags Defines which levels define the range. Use values of the enum
     *              collationMode.
     * \param collationIndices Defines which collations define the range it letters
     *                         have multiple collations.
     *
     * \see addWeightRange()
     * \see TableModel::addWeightRange()
     */
    void addWeightRangeSignal ( QStringList begin, QStringList end,
                                TableModel::CollationModes flags,
                                QList<int> collationIndices );
};

#endif /* TABLECONTROLLER_H_ */
