/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann           *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef TABLEMODEL_H_
#define TABLEMODEL_H_

#include "data/Letter.h"
#include <QAbstractTableModel>
#include <QList>
#include "gui/UCQBlockList.h"
#include <QProgressDialog>
#include <QApplication>
#include <QWidget>
#include <QHash>
#include <QListWidget>

/**
 * \class TableModel
 * \brief Data class for the table model.
 *
 * The TableModel class manages all Unicode letters and blocks, that are loaded
 * by the TableIO class.
 * It has two main purposes: it stores all letters that were loaded into the
 * application and it provides a table model for the visible table in a
 * QTableView. <br>
 * It consists of two lists and two hashmaps: items_, visibleItems_, blocks_
 * and blockitems_. <br>
 *
 * The items_ list contains objects of all letters, that were added to the
 * model with addLetter() or addLetterList().  <br>
 *
 * The visibleItems_ list contains pointers to those items in the complete list,
 * that should be visible in the TableView. Its contents can be changed with
 * the slots showXXX, addXXX, hideXXX and clearVisible(). Changes to the
 * visibleItems_ have direct influence on the table shown in the QTableView.<br>
 *
 * The blocks_ hashmap contains informations on blocks. The blocks name as key
 * and the blocks lower and upper bound as values. <br>
 *
 * The blockItems_ hashmap uses also the block names as keys, but its values are
 * pointers to those letters from the items list, which codepoints fall into
 * the blocks boundaries. New blocks can be added with the addBlock() method,
 * that will add the block information to blocks and populate the new key in
 * blockItems_ with pointers to items from the complete list. <br>
 *
 * For the use as a QTableModel this class inherits QAbstractTableModel and
 * implements its virtual methods rowCount(), columnCount() and data() and
 * provides two index() methods to get QModelIndex objects. Together with
 * UCQTableView, the visible table has a variable column count, which changes
 * when resize events occur. To make this possible it is necessary, that the
 * model reacts on resize signals from the view, which is usually unwanted in
 * a model-view architecture. It is not possible to achieve this behavior in
 * the view only, because in Qt the model defines how many columns the view
 * should have.<br>
 *
 * In addition to the features mentioned above, the TableModel helps to have a
 * QListWidget, to display the available blocks. Every time a block is added,
 * the signal blocksChanged() is emitted. (see UCQBlockList)
 *
 * \author Martin Burgemeister <Martin.Burgemeister@gmx.de>,
 *         Mathis Lindemann <MathisLindemann@gmx.net>
 * \version 1.0
 * \date 16 August 2010
 *
 * \see UCQBlockList
 * \see UCQTableView
 */
class TableModel: public QAbstractTableModel
{

Q_OBJECT

public:
    TableModel  ( QWidget* parent = 0 );

    int columnCount ( const QModelIndex& = QModelIndex () ) const;
    int rowCount    ( const QModelIndex& = QModelIndex () ) const;
    QVariant data   ( const QModelIndex&, int = Qt::DisplayRole ) const;

    void recalcLayout ();

    void addLetter     ( Letter,         bool = false );
    void addLetterList ( QList<Letter> , bool = false );

    void addBlock      ( QString, uint, uint end, bool = false );

    QModelIndex index ( int, int, const QModelIndex& = QModelIndex () ) const;
    QModelIndex index ( Letter* );

    QList<Letter*> getItems ();
    QList<Letter*> getVisibleItems ();
    QList<Letter*> getLetters ( QModelIndexList );


    Letter* getLetter ( QModelIndex );
    Letter* getVisibleLetter ( QModelIndex );

    UCQBlockList* getBlockList  ();
    QStringList   getBlockNames ();

    void setBlockList ( UCQBlockList* );

    bool contains ( Letter );
    bool hasTable ();
    bool hasBlocks ();
    bool hasVisibleTable ();
    /**
     * \brief Enum for the collations modes.
     */
    enum collationMode
    {
        COLLATION_STRING      =  0x1,
        COLLATION_LEVEL1      =  0x2,
        COLLATION_LEVEL2      =  0x4,
        COLLATION_LEVEL3      =  0x8,
        COLLATION_LEVEL4      = 0x10,
        COLLATION_VARIABLE    = 0x20,
        COLLATION_NOTVARIABLE = 0x40
    };

    Q_DECLARE_FLAGS(CollationModes, collationMode)

private:
    QList<Letter*> searchByLevel ( QList<Letter*> &, int, uint, uint,
                                   QList<int> );
    /**
     * \brief Contains all letters of the model.
     */
    QList<Letter> items_;
    /**
     * \brief Contains pointers to the items in items_ that are visible in the
     *        QTableView.
     */
    QList<Letter*> visibleItems_;
    /**
     * \brief Number of columns in the QTableView.
     */
    int viewColumns_;
    /**
     * \brief Stores the row count after last resize.
     */
    int oldRows_;
    /**
     * \brief Stores the column count after last resize.
     */
    int oldColumns_;
    /**
     * \brief Hashmap with blocknames pointers to the items of each block.
     */
    QHash<QString, Letter*> blockItems_;
    /**
     * \brief Hashmap with the blocknames and the bondaries of each block.
     */
    QHash<QString, uint> blocks_;
    /**
     * \brief QListWidget that displays the blocks.
     */
    UCQBlockList* blockList_;
    /**
     * \brief ParentWidget for progressbar.
     */
    QWidget* parentWidget_;
    /**
     * \brief Map of indexes in items_ for constant lookup time.
     *
     * The map contains the indices of the letters in items_. It is used,
     * because the lookup in QList depends on the number of entries.
     */
    QHash<Letter, int> indexMap;

public slots:
    void addWeightRange ( QStringList, QStringList, TableModel::CollationModes,
                          QList<int> );
    void clearTable   ();
    void clearModel   ();
    void clearVisible (bool = true);
    void clearBlocks  ();
    void showAll      ();
    void updateBlocks ();
//    void showLettersOfFont(const QFont&);

    void hideIndexes ( QModelIndexList );
    void showIndexes ( QModelIndexList );

    void hideRange ( uint, uint );
    void addRange  ( uint, uint );

    void addList ( QList<Letter*> );

    void showBlocks ( QStringList );
    void viewResized ( int );

signals:
    /**
     * \brief Signals a change in the blocks of the model.
     *
     * \param blocks List of blocks with the block names and the boundaries.
     */
    void blocksChanged ( QHash<QString, uint> blocks );
    /**
     * \brief Signals a warning, i.e. when a letter or block is replaced.
     *
     * \param warning The warnings text.
     */
    void newWarning ( QString warning );
    /**
     * \brief Signals a change in the layout that makes a resize in the view
     *        neccessary.
     */
    void viewResizeCells ();
    /**
     * \brief Signals a change in the visibleItems list.
     */
    void visibleChanged ();

};

Q_DECLARE_OPERATORS_FOR_FLAGS(TableModel::CollationModes)

#endif /* TABLEMODEL_H_ */
