/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann           *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef LUNA_H_
#define LUNA_H_

/**
 * \class Luna
 * \brief Makes calling non-static methods from Lua possible.
 *
 * This template provides methods to register and call non static methods
 * in the Lua C API.
 * The original source for this template can be found at
 * http://lua-users.org/wiki/LunaWrapper
 * inspired by
 * http://www.lua.org/notes/ltn005.html.
 *
 * \author Original Author: For Lua Version 5.0 Lenny Palozzi and for 5.1 Nornagon
 *         edited for this Software by
 *         Martin Burgemeister <Martin.Burgemeister@gmx.de>,
 *         Mathis Lindemann <MathisLindemann@gmx.net>
 * \version 1.0
 * \date 16 August 2010
 * \see LuaController
 */
template < class T > class Luna
{
public:
    /**
     * \brief Function map.
     *
     * This is a function map. Name holds the name of the function, that the
     * member function mfunc of the class will be registered to the in Lua
     * table.
     */
    struct RegType
    {
        /**
         * \brief Name of the method.
         */
        const char *name;
        /**
         * \brief Pointer on the method.
         */
        int(T::*mfunc) ( lua_State* );
    };

    /**
     * \brief Register a function in Lua.
     *
     * The Register method registers a method of the C-object in the Lua-object.
     * It registers the C-object in the form of a Lua-table constructed by the
     * constructor method. It also registers the gc_obj method in Lua, that
     * will delete the C-object.
     *
     * \param L The lua_State to use.
     * \see gc_obj()
     */
    static void Register ( lua_State* L )
    {
        lua_pushcfunction ( L, &Luna<T>::constructor );

        lua_setglobal     ( L, T::className );
        luaL_newmetatable ( L, T::className );

        lua_pushstring    ( L, "__gc" );
        lua_pushcfunction ( L, &Luna<T>::gc_obj );

        lua_settable      ( L, -3 );
    }

private:
    /**
     * \brief Creates a Table for the object with all methods.
     *
     * This method creates the Lua-table that represents the C-object
     * in Lua. The method registers every member function of the class to the
     * Lua-table and returns the table to the Lua interpreter. The pointer to
     * the object is stored in the table on index 0 as userdata. All functions
     * are stored as a closure value.
     *
     * \param L The lua_State to use.
     * \see thunk()
     */
    static int constructor ( lua_State* L )
    {
        // new T
        T* obj = new T ( L );

        if(obj->isNull()){
            delete obj;
            return 0;
        }

        if(!obj->isController()){
        	delete obj;
        	return 0;
        }

        // New table object
        lua_newtable   ( L );
        // Userdata obj at index 0
        lua_pushnumber ( L, 0 );

        T** a = (T**) lua_newuserdata ( L, sizeof(T*) );
        *a = obj;
        luaL_getmetatable ( L, T::className );
        lua_setmetatable  ( L, -2 );
        // table[0] = obj;
        lua_settable      ( L, -3 );

        // Register the member functions
        for ( int i = 0; T::Register[ i ].name; i++ )
        {
            lua_pushstring ( L, T::Register[ i ].name );
            lua_pushnumber ( L, i );
            lua_pushcclosure ( L, &Luna<T>::thunk, 1 );
            lua_settable   ( L, -3 );
        }
        // Return the table object
        return 1;
    }

    /**
     * \brief "Translates" between the Lua function-call and the C++-method call.
     *
     * All member functions of the class are registered in Lua with the
     * function. The function pointer is indexed in the member function map
     * by the name of the function.
     *
     * \param L The lua_State to use.
     */
    static int thunk ( lua_State* L )
    {
        int i = (int) lua_tonumber ( L, lua_upvalueindex ( 1 ) );
        // userdata object at index 0
        lua_pushnumber ( L, 0 );
        lua_gettable   ( L, 1 );

        T** obj = static_cast<T**> ( luaL_checkudata ( L, -1, T::className ) );
        lua_remove ( L, -1 );
        return ( ( *obj )->*( T::Register[ i ].mfunc ) ) ( L );
    }

    /**
     * \brief Garbage collector for the object.
     *
     *  This function deletes the object. This function is registered
     *  with the name '__gc' in the Lua-table and will be called, if
     *  the garbage collector is run.
     *
     *  \param L The lua_State to use.
     */
    static int gc_obj ( lua_State* L )
    {
        T** obj = static_cast<T**> ( luaL_checkudata ( L, -1, T::className ) );
        delete ( *obj );
        return 0;
    }
protected:
    /**
     * \brief Hidden default constructor.
     */
    Luna (); // Hide default constructor
};

#endif /* LUATABLE_H_ */
