/***************************************************************************
 *   Copyright (C) 2010 by Martin Burgemeister, Mathis Lindemann           *
 *     																	   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "mainwindow.h"

/**
 * \brief Constructs the GUI and establishes many required connections between
 *        elements of the GUI.
 *
 * \param parent Pointer to the parent QWidget.
 */
MainWindow::MainWindow ( QWidget* parent ) :
    QMainWindow ( parent ), executeLuaThread_(lua_,UCQExecuteLuaThread::EXECUTE_FILE)

{
    ui_.setupUi ( this );

    setWindowFlags ( windowFlags () | Qt::CustomizeWindowHint
                                    | Qt::WindowMinMaxButtonsHint
                                    | Qt::WindowSystemMenuHint );
    setWindowIcon ( QIcon ( ":icons/uc-book.svg" ) );
    setWindowTitle ( "Unicode Viewer" );

    // detailsbars
    ui_.utf8DetailsBar->setUtf8    ( true );
    ui_.utf8DetailsBar->setVisible ( false );
    ui_.utf32Box->setVisible       ( false );

    // initialize objects
    tableModel_  = new TableModel       ( this );
    sc_          = new SearchController ( this, tableModel_, ui_.tableView );
    details_     = new UCQDetails       ( this, tableModel_ );
    ucqsd_       = new UCQSearchDialog  ( this, sc_ );
    helpBrowser_ = 0;
    working_     = false;

    initMenu ();
    initStatusBar ();
    // restore state of application before last close
    readSettings ();

    tableModel_->setBlockList ( ui_.blockList );

    ui_.tableView->setModel           ( tableModel_ );
    ui_.detailsBar->setTableModel     ( tableModel_ );
    ui_.utf8DetailsBar->setTableModel ( tableModel_ );
//    ui_.fontList->setTableModel       ( tableModel_ );
//    ui_.fontList->setFontBar(ui_.toolBarFont);

    updateStates ();



    qRegisterMetaType<QItemSelection> ( "QItemSelection" );
    qRegisterMetaType<QTextCursor> ( "QTextCursor" );

    connect ( &fileSystemWatcher, SIGNAL ( fileChanged ( QString ) ),
              this,               SLOT   ( fileChanged ( QString ) ));

    // <<<<<<<<<<<<<<<< ToolbarFont >>>>>>>>>>>>>>>>>
    connect ( ui_.toolBarFont, SIGNAL ( currentFontChanged ( const QFont & )),
              ui_.tableView,   SLOT   ( changeFont         ( const QFont & )));

    connect ( ui_.toolBarFont, SIGNAL ( currentSizeChanged( const QString & )),
              ui_.tableView, SLOT   ( changeFontSize    ( const QString & ) ));

    connect ( ui_.toolBarFont, SIGNAL ( bold_toggled ( bool ) ),
              ui_.tableView,   SLOT   ( setBold      ( bool ) ));

    connect ( ui_.toolBarFont, SIGNAL ( italic_toggled ( bool ) ),
              ui_.tableView,   SLOT   ( setItalic      ( bool ) ));

    connect ( ui_.toolBarFont, SIGNAL( currentFontChanged ( const QFont & ) ),
              details_,        SLOT  ( setFont            ( const QFont & ) ));

    ui_.toolBarFont->initFont ();

    // <<<<<<<<<<<<<<< TableView >>>>>>>>>>>>>>>>>>>>>
    connect ( ui_.tableView,  SIGNAL ( newSelection( QModelIndexList ) ),
              ui_.detailsBar, SLOT   ( changeStatus( QModelIndexList ) ));

    connect ( ui_.tableView,      SIGNAL ( newSelection( QModelIndexList ) ),
              ui_.utf8DetailsBar, SLOT   ( changeStatus( QModelIndexList ) ));

    connect ( ui_.tableView, SIGNAL ( printSelection( QList<Letter*> ) ),
              ui_.console,   SLOT   ( printSelection( QList<Letter*> ) ));

    connect ( ui_.tableView, SIGNAL ( doubleClicked ( Letter * ) ),
              ui_.console,   SLOT   ( printLetter   ( Letter * ) ));

    connect ( ui_.tableView, SIGNAL ( showDetails ( QMouseEvent* ) ),
              details_,      SLOT   ( changeStatus( QMouseEvent* ) ));

    connect ( ui_.tableView, SIGNAL ( closeDetails() ),
              details_,      SLOT   ( close ()       ));

    connect ( ui_.tableView, SIGNAL ( moveDetails ( int, int ) ),
              details_,      SLOT   ( move        ( int, int ) ));

    connect ( ui_.tableView, SIGNAL ( entered ( const QModelIndex& ) ),
              details_,      SLOT   ( changeLetter(   QModelIndex  ) ));

    connect ( ui_.tableView, SIGNAL ( resized     ( int ) ),
              tableModel_,   SLOT   ( viewResized ( int ) ));

    connect ( ui_.actionZoomIn, SIGNAL ( triggered () ),
              ui_.tableView,    SLOT   ( zoomIn    () ));

    connect ( ui_.actionZoomOut, SIGNAL ( triggered () ),
              ui_.tableView,     SLOT   ( zoomOut   () ));

    // <<<<<<<<<<<<<< tableToolBar >>>>>>>>>>>>>>>>>>>>>
    connect ( ui_.tableToolBar, SIGNAL ( zoomInPressed () ),
              ui_.tableView,    SLOT   ( zoomIn        () ));

    connect ( ui_.tableToolBar, SIGNAL ( zoomOutPressed () ),
              ui_.tableView,    SLOT   ( zoomOut        () ));

    connect ( ui_.tableToolBar, SIGNAL ( searchPressed             () ),
              this,             SLOT   ( on_actionSearch_triggered () ));

    connect ( ui_.tableToolBar, SIGNAL ( addBlockPressed () ),
              this,             SLOT   ( showBlockDialog () ));

    connect ( ui_.tableToolBar, SIGNAL ( clearPressed () ),
              this,             SLOT   ( clearTable   () ));

    connect ( ui_.tableToolBar, SIGNAL ( cutSelectionPressed () ),
              this,             SLOT   ( cutSelectionAction  () ));

    connect ( ui_.tableToolBar, SIGNAL ( cropSelectionPressed () ),
              this,             SLOT   ( cropSelectionAction  () ));

    // <<<<<<<<<<<<< toolBarUC >>>>>>>>>>>>>>>>>>>>>>>
    connect ( ui_.toolBarUc, SIGNAL ( loadTablePressed              () ),
              this,          SLOT   ( on_actionUcTableLoad_triggered() ));

    connect ( ui_.toolBarUc, SIGNAL ( saveTablePressed               () ),
              this,          SLOT   ( on_actionUcTableSave_triggered () ));

    connect ( ui_.toolBarUc, SIGNAL ( loadBlocksTablePressed            () ),
              this,          SLOT   ( on_actionBlocksTableLoad_triggered() ));

    // <<<<<<<<<<<<<< tableModel >>>>>>>>>>>>>>>>>>>>>
    connect ( tableModel_,   SIGNAL ( blocksChanged   ( QHash<QString, uint>)),
              ui_.blockList, SLOT  ( contentsChanged ( QHash<QString, uint>)));

    connect ( tableModel_,    SIGNAL ( viewResizeCells () ),
              ui_.tableView,  SLOT   ( resizeCells     () ));

    connect ( tableModel_, SIGNAL ( visibleChanged () ),
              this,        SLOT   ( updateStates   () ));

    // <<<<<<<<<<<<<< blockList >>>>>>>>>>>>>>>>>>>>>>
    connect ( ui_.blockList, SIGNAL ( currentBlockChanged ( QStringList ) ),
              tableModel_,   SLOT   ( showBlocks          ( QStringList ) ));

    connect ( ui_.blockList, SIGNAL ( blockChanged    () ),
              ui_.tableView, SLOT   ( clearSelections () ));

    connect ( ui_.blockList, SIGNAL ( blockChanged () ),
              this,          SLOT   ( updateStates () ));

    connect ( this, SIGNAL ( sortBlocks(UCQBlockList::SortingOrders) ),
              ui_.blockList,       SLOT( sortBlocks( UCQBlockList::SortingOrders )));

}

/**
 * \brief Stops any running Lua execution and closes Lua to free all Lua
 *        resources.
 *
 * \see stopExecution()
 * \see LuaController::stop()
 * \see LuaController::close()
 */
MainWindow::~MainWindow ()
{
//    emit stopExecution ();
//    executeLuaThread_.quit ();
//    lua_->close ();
//    ui_.console->stopExec();
//    ui_.console->~UCQConsole();
    lua_->close ();
    delete sortAscIcon_;
    delete sortDescIcon_;
}

/********************************************************************
 * Initialization                                                   *
 ********************************************************************/
/**
 * \brief Initializes the UCQStatusBar status bar with a new UCQWarnings
 *        object.
 *
 * \see UCQWarnings::UCQWarnings()
 * \see UCQStatusBar::UCQStatusBar()
 */
void MainWindow::initStatusBar ()
{
    warningsDialog_ = new UCQWarnings ( this );

    connect ( tableModel_,     SIGNAL ( newWarning ( QString ) ),
              warningsDialog_, SLOT   ( addWarning ( QString ) ));

    warningsDialog_->hide ();

    // exchange status bar created by the designer
    statusBar_ = new UCQStatusBar ( this, warningsDialog_ );
    setStatusBar ( statusBar_ );
}

/**
 * \brief Initializes the toggle checkboxes for the console and the blocklist
 *        in the "View" menu and adds the recent files submenu.
 *
 * \see openRecentFile()
 * \see openRecentBlock()
 */
void MainWindow::initMenu ()
{
    // checkbox in menu to hide console
    QAction* consoleToggle = ui_.dockWidgetConsole->toggleViewAction ();
    consoleToggle->setShortcut ( QKeySequence ( tr ( "Alt+C" ) ) );
    ui_.menuShow_Views->addAction ( consoleToggle );

    // checkbox in menu to hide blocklist
    QAction* scriptsToggle = ui_.dockWidgetScripts->toggleViewAction ();
    scriptsToggle->setShortcut ( QKeySequence ( tr ( "Alt+B" ) ) );
    ui_.menuShow_Views->addAction ( scriptsToggle );

    // initialize recent files submenus
    for ( int i = 0; i < MaxRecentFiles; ++i )
    {
        recentFileActions_[ i ] = new QAction ( this );
        recentFileActions_[ i ]->setVisible ( false );
        recentBlocksActions_[ i ] = new QAction ( this );
        recentBlocksActions_[ i ]->setVisible ( false );

        connect ( recentFileActions_[ i ], SIGNAL ( triggered      () ),
                  this,                    SLOT   ( openRecentFile () ));

        connect ( recentBlocksActions_[ i ], SIGNAL ( triggered       () ),
                  this,                      SLOT   ( openRecentBlock () ));

        ui_.menuLatestUC->addAction     ( recentFileActions_[ i ] );
        ui_.menuLatestBlocks->addAction ( recentBlocksActions_[ i ] );
    }

    // group for block sorting
    blockSortingActions = new QActionGroup(this);
    blockSortingActions->setExclusive(true);
    blockSortingActions->addAction(ui_.actionBlocksNameAsc);
    blockSortingActions->addAction(ui_.actionBlocksNameDesc);
    blockSortingActions->addAction(ui_.actionBlocksCodeAsc);
    blockSortingActions->addAction(ui_.actionBlocksCodeDesc);

    // group for table sorting attribute
    tableSortingActionsAttributes = new QActionGroup(this);
    tableSortingActionsAttributes->setExclusive(true);
    tableSortingActionsAttributes->addAction(ui_.actionSortDefault);
    tableSortingActionsAttributes->addAction(ui_.actionSortCodepoint);
    tableSortingActionsAttributes->addAction(ui_.actionSortCollation);
    tableSortingActionsAttributes->addAction(ui_.actionSortComment);

    // group for table sorting order
    tableSortingActionsOrders = new QActionGroup(this);
    tableSortingActionsOrders->setExclusive(true);
    tableSortingActionsOrders->addAction(ui_.actionAscending);
    tableSortingActionsOrders->addAction(ui_.actionDescending);

    sortAscIcon_ = new QIcon( ":icons/sort-asc.svg" );
    sortDescIcon_ = new QIcon( ":icons/sort-desc.svg" );
    ui_.sortOrderButton->setIcon(*sortAscIcon_);
    ui_.sortOrderButton->setToolTip(tr("Ascending order"));
    ascending_ = true;
}

/**
 * \brief Initialize Lua.
 *
 * This function initializes the LuaController, executes the initial Lua-files
 * and registers the required objects to Lua.
 *
 * \see LuaController::Register
 * \see LuaController::LuaController()
 * \see LuaController::getCompletionList()
 * \see LuaController::stop()
 * \see LuaController::doFile()
 * \see TableController::TableController()
 * \see SortController::SortController()
 * \see UCQConsole::error()
 * \see UCQConsole::getCompletionList()
 * \see UCQConsole::executionReady()
 * \see stopExecution()

 */
void MainWindow::initLua ()
{
    lua_          = new LuaController   ( this );
    tableControl_ = new TableController ( this, tableModel_ );
    sortControl_  = new SortController  ( this, tableModel_ );

//    connect ( tableModel_, SIGNAL ( visibleChanged () ),
//                sortControl_,        SLOT   ( restoreDefault   () ), Qt::QueuedConnection);

    executeLuaThread_.setLuaController(lua_);

    connect ( lua_,        SIGNAL ( error ( QString ) ),
              ui_.console, SLOT   ( error    ( QString ) ));

    connect ( lua_, SIGNAL ( printSignal ( QString ) ),
    		ui_.console, SLOT   ( append      ( QString ) ),
              Qt::BlockingQueuedConnection
            );

    connect ( ui_.console, SIGNAL ( stopExecution() ),
              lua_, SLOT   ( stop() )
            );

    connect ( ui_.console, SIGNAL
                       ( getCompletionList ( const QString&, QStringList& ) ),
              lua_,        SLOT
                       ( getCompletionList ( const QString&, QStringList& ) ));
    connect ( &executeLuaThread_, SIGNAL
                       ( running (bool) ),
              ui_.pushButtonLoadLua,        SLOT
                       ( setDisabled(bool) ));

    connect ( this, SIGNAL ( stopExecution () ),
              lua_, SLOT   ( stop          () ), Qt::QueuedConnection);

    connect (  ui_.pushButtonStop, SIGNAL ( clicked () ),
                 lua_, SLOT   ( stop() ), Qt::QueuedConnection);
    connect ( &executeLuaThread_, SIGNAL
                       ( running (bool) ),
              ui_.pushButtonStop,        SLOT
                       ( setEnabled(bool) ));

    // Register the Objects
    lua_->Register<LuaSearch, SearchController> ( sc_,        "searchcontrol");
    lua_->Register<LuaTable, TableController> ( tableControl_, "tablecontrol");
    lua_->Register<LuaOutput, LuaController>  ( lua_,          "controller"  );
    lua_->Register<LuaSort, SortController>   ( sortControl_,  "sortcontrol" );


    // find files (platform dependent)
#if defined(LINUX)
    QDir dir(RESOURCEPATH);
#else
    QDir dir ( QApplication::applicationDirPath () );

#if defined(Q_OS_WIN)
    if (dir.dirName().toLower() == "debug"
            || dir.dirName().toLower() == "release")
    dir.cdUp();

#elif defined(Q_OS_MAC)
    if (dir.dirName() == "MacOS")
    {
        dir.cdUp();
        dir.cd("Resources");
    }

#endif
    dir.cd ( "lua" );
#endif

    // load localization tables
    QString locale = dir.absolutePath () + QDir::separator ()
            + QLocale::system ().name () + ".lua";
    if ( !QFileInfo ( locale ).exists () )
        locale = dir.absolutePath () + QDir::separator () + "en_EN.lua";

    // execute init.lua
    lua_->doFile ( locale );
    lua_->doFile ( dir.absolutePath () + QDir::separator () + "init.lua" );

#if defined(LINUX)
    dir = QDir(QDir::homePath ()+"/.UnicodeViewer");
#endif

    // if user.lua not found, create it
    QString userlua = dir.absolutePath () + QDir::separator () + "user.lua";
    if ( !dir.exists () )
    {
        dir.mkdir ( dir.absolutePath () );
    }

    bool justCreated = false;
    if ( !QFileInfo ( userlua ).exists () )
    {
        QFile user ( dir.absolutePath () + QDir::separator () + "user.lua" );
        user.open ( QIODevice::ReadWrite );
        user.close ();
        justCreated = true;
    }

    ui_.console->setExecuteThread(&executeLuaThread_);

    // user.lua was not just created and is not empty, execute it
    if ( !justCreated && QFileInfo ( userlua ).size () > 0 )
    {
    	executeLuaThread_.execute ( dir.absolutePath () + QDir::separator ()
                + "user.lua" );
    } else
    {
        ui_.console->print ( "" );
        ui_.console->displayPrompt ();
    }
    connect ( lua_, SIGNAL ( startExecution  () ),
               ui_.console, SLOT   ( executionInit () ),
               Qt::BlockingQueuedConnection
             );

    connect ( lua_, SIGNAL ( endExecution  () ),
               ui_.console, SLOT   ( executionReady () ),
               Qt::BlockingQueuedConnection
             );
}

/********************************************************************
 * MainWindow Actions                                               *
 ********************************************************************/
// <<<<<<<<<<<<<<<<<<<<<< Table Operations >>>>>>>>>>>>>>>>>>>>>>>>>>
/**
 * \brief Connects the clear table action with the TableModel.
 *
 * \see TableModel::clearVisible()
 */
void MainWindow::clearTable ()
{
    tableModel_->clearVisible ();
}

/**
 * \brief Connects the crop selection menu item with the TableModel.
 *
 * \see TableModel::showIndexes()
 */
void MainWindow::cropSelectionAction ()
{
    tableModel_->showIndexes ( ui_.tableView->getSelectedIndexes () );
}

/**
 * \brief Connects the cut selection action with the TableModel.
 *
 * \see TableModel::hideIndexes()
 */
void MainWindow::cutSelectionAction ()
{
    tableModel_->hideIndexes ( ui_.tableView->getSelectedIndexes () );
}

/**
 * \brief Slot for adding block tables.
 *
 * \see loadBlockTable()
 */
void MainWindow::on_actionAddBlocksTable_triggered ()
{
    if ( curBlocksDir_.isEmpty () )
    {
        curBlocksDir_ = QDir::toNativeSeparators ( QDir::homePath () );
    }
    QString file = QFileDialog::getOpenFileName ( this,
                                                  tr ( "Add Blocks File" ),
                                                  curBlocksDir_,
                                                  tr ( "Textfile (*.txt)" ) );
    if ( !file.isEmpty () && QFile::exists ( file ) )
    {
        loadBlockTable ( file, true );
    }
}

/**
 * \brief Slot for adding Unicode tables.
 *
 * \see loadTable()
 */
void MainWindow::on_actionAddUcTable_triggered ()
{
    if ( curTableDir_.isEmpty () )
    {
        curTableDir_ = QDir::toNativeSeparators ( QDir::homePath () );
    }
    QString file = QFileDialog::getOpenFileName ( this,
                                                  tr ( "Add Unicode File" ),
                                                  curTableDir_,
                                                  tr ( "Textfile (*.txt)" ) );
    if ( !file.isEmpty () && QFile::exists ( file ) )
    {
        loadTable ( file, true );
    }
}

/**
 * \brief Slot for loading block tables.
 *
 * \see loadBlockTable()
 */
void MainWindow::on_actionBlocksTableLoad_triggered ()
{
    if ( curBlocksDir_.isEmpty () )
    {
        curBlocksDir_ = QDir::toNativeSeparators ( QDir::homePath () );
    }
    QString file = QFileDialog::getOpenFileName ( this,
                                                  tr ( "Open Blocks File" ),
                                                  curBlocksDir_,
                                                  tr ( "Textfile (*.txt)" ) );
    if ( !file.isEmpty () && QFile::exists ( file ) )
    {
        loadBlockTable ( file, false );
    }
}

/**
 * \brief Slot that resets the application.
 *
 * \see TableModel::clearTable()
 * \see TableModel::clearBlocks()
 * \see UCQWarnings::clear()
 * \see UCQConsole::clearConsole()
 * \see SearchController::clear()
 */
void MainWindow::on_actionReset_triggered ()
{
    setCurrentFile ( "" );
    tableModel_->clearTable ();
    tableModel_->clearBlocks ();
    warningsDialog_->clear ();
    sc_->clear ();
    ui_.console->clearConsole ();
    setWindowTitle ( "Unicode Viewer" );
    if ( !fileSystemWatcher.files ().isEmpty () )
    {
        fileSystemWatcher.removePaths ( fileSystemWatcher.files () );
    }
}

/**
 * \brief Slot for loading Unicode tables.
 *
 * \see loadTable()
 * \see loadBlockTable()
 */
void MainWindow::on_actionUcTableLoad_triggered ()
{
    if ( curTableDir_.isEmpty () )
    {
        curTableDir_ = QDir::toNativeSeparators ( QDir::homePath () );
    }
    QString file = QFileDialog::getOpenFileName ( this,
                                                  tr ( "Open Unicode File" ),
                                                  curTableDir_,
                                                  tr ( "Textfile (*.txt)" ) );
    if ( !file.isEmpty () && QFile::exists ( file ) )
    {
        loadTable ( file, false );
        QString blocksFile = QFileInfo ( file ).path () + QDir::separator ()
                + "Blocks.txt";
        if ( QFile::exists ( blocksFile ) && !tableModel_->hasBlocks () )
            loadBlockTable ( blocksFile, false );

    }
}

/**
 * \brief Slot for exporting the visible table.
 *
 * \see TableIO::writeVisibleTable()
 */
void MainWindow::on_actionUcTableSave_triggered ()
{
    if ( curOutputDir_.isEmpty () )
    {
        curOutputDir_ = QDir::toNativeSeparators ( QDir::homePath () );
    }
    QString file = QFileDialog::getSaveFileName ( this, tr ( "Save Table" ),
                                                  curOutputDir_,
                                                  tr ( "Textfile (*.txt)" ) );
    if ( !file.isEmpty () )
    {
        bool ok = TableIO::writeVisibleTable ( file, *tableModel_ );
        if ( !ok )
        {
            statusBar_->showMessage ( tr ( "Saving failed." ), 10000 );
        }
        statusBar_->showMessage ( tr ( "Table successfully saved in" ) + " "
                + file + ".", 10000 );
        curOutputDir_
                = QDir::toNativeSeparators ( QFileInfo ( file ).absolutePath () );
    }
}

/**
 * \brief Opens the dialog to create new blocks.
 *
 */
void MainWindow::showBlockDialog ()
{
    UCQBlockDialog * bd = new UCQBlockDialog ( this, tableModel_ );
    bd->show ();
}

// <<<<<<<<<<<<<<<<<<<<<<<<<<< Search >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

/**
 * \brief Opens the search window.
 *
 * \see UCQSearchDialog::show()
 */
void MainWindow::on_actionSearch_triggered ()
{
    ucqsd_->show ();
}
/**
 * \brief Slot for selecting the next search hit.
 *
 * \see SearchController::next()
 */
void MainWindow::on_actionSearchNext_triggered ()
{
    if ( sc_->isEmpty () )
    {
        emit ui_.actionSearch->trigger ();
    } else
    {
        sc_->next ();
    }
}
/**
 * \brief Slot for selecting the previous search hit.
 *
 * \see SearchController::previous()
 */
void MainWindow::on_actionSearchPrevious_triggered ()
{
    if ( sc_->isEmpty () )
    {
        emit ui_.actionSearch->trigger ();
    } else
    {
        sc_->previous ();
    }
}

// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<< LUA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

/**
 * \brief Shows a QFileDialog and executes a Lua file.
 *
 * After the Load Lua Button was clicked, the dialog is opened and
 * the chosen file executed.
 *
 * \see UCQExecuteFileThread::setLuaController()
 * \see UCQExecuteFileThread::setString()
 * \see UCQExecuteFileThread::run()
 * \see UCQConsole::executionInit()
 */
void MainWindow::on_pushButtonLoadLua_clicked ()
{

    if ( curLuaDir_.isEmpty () )
    {
        curLuaDir_ = QDir::toNativeSeparators ( QDir::homePath () );
    }
    QString tmp = QFileDialog::getOpenFileName ( this, tr ( "Open Lua File" ),
                                                 curLuaDir_,
                                                 tr ( "Lua Script (*.lua)" ) );

    if ( tmp.isEmpty () )
    {
        return;
    }

    ui_.console->insertPlainText ( "Load file: \"" + tmp + "\"" );
    curLuaDir_ = QDir::toNativeSeparators ( QFileInfo ( tmp ).absolutePath () );
    executeLuaThread_.setexecutionMode(UCQExecuteLuaThread::EXECUTE_FILE);
    executeLuaThread_.execute(tmp);
}

// <<<<<<<<<<<<<<<<<<<<<<<<<<<<< Help >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
/**
 * \brief Opens the help browser.
 */
void MainWindow::on_actionHelp_triggered ()
{
    if ( helpBrowser_ == 0 )
    {
        helpBrowser_ = new UCQHelpBrowser ( this );
    }
    helpBrowser_->show ();
}

/**
 * \brief Opens the about text in the UCQHelpBrowser.
 *
 * \see UCQHelpBrowser::jumpTo()
 */
void MainWindow::on_actionAbout_triggered ()
{
    on_actionHelp_triggered ();
    if ( helpBrowser_ == 0 )
    {
        helpBrowser_ = new UCQHelpBrowser ( this );
    }
    helpBrowser_->jumpTo ( "about.html" );
}
/********************************************************************
 * Settings			                                                *
 ********************************************************************/
/**
 * \brief Writes the settings and recent files.
 */
void MainWindow::writeSettings ()
{
    QSettings settings ( "Unicode Viewer", "Unicode Viewer" );
    settings.setValue ( "geometry", saveGeometry () );
    settings.setValue ( "recentFiles", recentFiles_ );
    settings.setValue ( "recentBlocks", recentBlocks_ );
    settings.setValue ( "curTableDir", curTableDir_ );
    settings.setValue ( "curBlocksDir", curBlocksDir_ );
    settings.setValue ( "curOutputDir", curOutputDir_ );
    settings.setValue ( "curLuaDir", curLuaDir_ );
    QString temp = "";
    if ( !curFile_.isEmpty () )
    {
        temp = curFile_.last ();
    }
    settings.setValue ( "curFile", temp );
    settings.setValue ( "curBlock", curBlock_ );
}

/**
 * \brief Reads the settings and recent files.
 */
void MainWindow::readSettings ()
{
    QSettings settings ( "Unicode Viewer", "Unicode Viewer" );
    restoreGeometry ( settings.value ( "geometry" ).toByteArray () );
    recentFiles_ = settings.value ( "recentFiles" ).toStringList ();
    recentBlocks_ = settings.value ( "recentBlocks" ).toStringList ();
    curTableDir_ = settings.value ( "curTableDir" ).toString ();
    curBlocksDir_ = settings.value ( "curBlocksDir" ).toString ();
    curOutputDir_ = settings.value ( "curOutputDir" ).toString ();
    curLuaDir_ = settings.value ( "curLuaDir" ).toString ();
    curFile_.append ( settings.value ( "curFile" ).toString () );
    curBlock_ = settings.value ( "curBlock" ).toString ();
    updateRecentFileActions ();
    updateRecentBlocksActions ();
}

/********************************************************************
 * File management		                                            *
 ********************************************************************/

/**
 * \brief Loads the block file.
 *
 * \param filename Name of the file.
 * \param keepOld If true, replacement of blocks will cause new warnings.
 *
 * \see TableModel::clearBlocks()
 * \see TableIO::readBlockTable()
 * \see UCQBlockList::selectBlocks()
 */
void MainWindow::loadBlockTable ( QString filename, bool keepOld )
{
    if ( filename.isEmpty () )
    {
        return;
    }
    if ( !keepOld )
    {
        tableModel_->clearBlocks ();

    }
    statusBar_->showMessage ( tr ( "Loading" ) + " " + filename + tr ( "..." ),
                              10000 );

    bool readOk = TableIO::readBlockTable ( filename, *tableModel_, keepOld );
    if ( !readOk )
    {
        statusBar_->showMessage ( tr ( "Failed to load" ) + filename );
        return;
    }

    setCurrentBlock ( filename );
    ui_.blockList->selectBlocks ( QStringList ( tr ( "Show all" ) ), true );
    statusBar_->showMessage ( filename + " " + tr ( "loaded successfully." ),
                              5000 );
}

/**
 * \brief Loads a Unicode table.
 *
 * \param filename Name of the file.
 * \param keepOld If true, the table is appended to the current.
 *
 * \see TableModel::clearTable()
 * \see TableIO::readTable()
 * \see TableModel::showAll()
 */
void MainWindow::loadTable ( QString filename, bool keepOld )
{
    if ( filename.isEmpty () )
    {
        return;
    }
    if ( !keepOld )
    {
        tableModel_->clearTable ();
        if ( !curFile_.isEmpty () )
        {
            fileSystemWatcher.removePaths ( curFile_ );
            curFile_.clear ();
        }
    }

    statusBar_->showMessage ( tr ( "Loading" ) + " " + filename + tr ( "..." ),
                              10000 );

    bool readOk = TableIO::readTable ( filename, *tableModel_, keepOld );
    if ( !readOk )
    {
        statusBar_->showMessage ( tr ( "Failed to load" ) + " " + filename );
        return;
    }

    setCurrentFile ( filename );
    if ( !fileSystemWatcher.files ().contains ( filename ) )
    {
        fileSystemWatcher.addPath ( filename );
    }
    tableModel_->showAll ();
    statusBar_->showMessage ( filename + " " + tr ( "loaded successfully." ),
                              5000 );
    setWindowTitle ( "Unicode Viewer - " + filename );
}

/**
 * \brief Loads a recent block table.
 *
 * \see loadBlockTable()
 */
void MainWindow::openRecentBlock ()
{
    QAction* action = qobject_cast<QAction *> ( sender () );
    if ( action )
    {
        QString file = action->text ();
        if ( !file.isEmpty () && QFile::exists ( file ) )
        {
            loadBlockTable ( file, false );
        }
    }
}

/**
 * \brief Loads a recent Unicode table.
 *
 * \see loadTable()
 * \see loadBlockTable()
 */
void MainWindow::openRecentFile ()
{
    QAction* action = qobject_cast<QAction *> ( sender () );
    if ( action )
    {
        QString file = action->text ();
        if ( !file.isEmpty () && QFile::exists ( file ) )
        {
            loadTable ( file, false );
            QString blocksFile = QFileInfo ( file ).path ()
                    + QDir::separator () + "Blocks.txt";
            if ( QFile::exists ( blocksFile ) && !tableModel_->hasBlocks () )
            {
                loadBlockTable ( blocksFile, false );
            }
        }
    }
}

/**
 * \brief Reloads the files when they are changed on disk and the user agrees
 *        to reload.
 *
 * \see TableIO::readTable()
 * \see TableModel::clearTable()
 * \see TableModel::showAll()
 */
void MainWindow::reloadFiles ()
{
    if ( tableModel_ == 0 || curFile_.isEmpty () )
        return;
    tableModel_->clearTable ();
    bool readOk = false;
    readOk = TableIO::readTable ( curFile_.at ( 0 ), *tableModel_, false );

    for ( int i = 1; i < curFile_.size (); ++i )
    {
        if ( !readOk )
        {
            statusBar_->showMessage (
                    tr ( "Error reloading one or more files." ) );
            return;
        }
        if ( QFile::exists ( curFile_.at ( i ) ) )
        {
            readOk
                    = TableIO::readTable ( curFile_.at ( i ), *tableModel_,
                                           true );
        } else
        {
            statusBar_->showMessage ( tr ( "Cant find:" ) + " "
                    + curFile_.at ( i ) + ". " + tr ( "Skipped." ) );
        }
    }
    tableModel_->showAll ();
}

/**
 * \brief Sets the current block file.
 *
 * \param filename Path to current file.
 * \see updateRecentBlocksActions()
 */
void MainWindow::setCurrentBlock ( const QString& filename )
{
    curBlock_ = filename;
    curBlocksDir_
            = QDir::toNativeSeparators ( QFileInfo ( filename ).absolutePath () );
    setWindowModified ( false );
    QString shownName = "Untitled";
    if ( !curBlock_.isEmpty () )
    {
        shownName = QFileInfo ( curBlock_ ).fileName ();
        recentBlocks_.removeAll ( curBlock_ );
        recentBlocks_.prepend ( curBlock_ );
        updateRecentBlocksActions ();
    }
}

/**
 * \brief Sets the current Unicode table file.
 *
 * \param filename Path to current file.
 * \see updateRecentBlocksActions()
 */
void MainWindow::setCurrentFile ( const QString& filename )
{
    if ( filename.isEmpty () )
        return;
    curFile_.append ( filename );
    curTableDir_
            = QDir::toNativeSeparators ( QFileInfo ( filename ).absolutePath () );
    setWindowModified ( false );
    QString shownName = "Untitled";
    if ( !curFile_.isEmpty () && !curFile_.last ().isEmpty () )
    {
        shownName = QFileInfo ( curFile_.last () ).fileName ();
        recentFiles_.removeAll ( curFile_.last () );
        recentFiles_.prepend ( curFile_.last () );
        updateRecentFileActions ();
    }
}

/**
 * \brief Updates the recent blocks actions.
 */
void MainWindow::updateRecentBlocksActions ()
{
    QMutableStringListIterator i ( recentBlocks_ );
    while ( i.hasNext () )
    {
        if ( !QFile::exists ( i.next () ) )
        {
            i.remove ();
        }
    }

    for ( int j = 0; j < MaxRecentFiles; ++j )
    {
        if ( j < recentBlocks_.count () )
        {
            recentBlocksActions_[ j ]->setText ( recentBlocks_[ j ] );
            recentBlocksActions_[ j ]->setVisible ( true );
        } else
        {
            recentBlocksActions_[ j ]->setVisible ( false );
        }
    }
}

/**
 * \brief Updates the recent Unicode files actions.
 */
void MainWindow::updateRecentFileActions ()
{
    QMutableStringListIterator i ( recentFiles_ );
    while ( i.hasNext () )
    {
        if ( !QFile::exists ( i.next () ) )
        {
            i.remove ();
        }
    }
    for ( int j = 0; j < MaxRecentFiles; ++j )
    {
        if ( j < recentFiles_.count () )
        {
            recentFileActions_[ j ]->setText ( recentFiles_[ j ] );
            recentFileActions_[ j ]->setVisible ( true );
        } else
        {
            recentFileActions_[ j ]->setVisible ( false );
        }
    }
}

/********************************************************************
 * Programm state			                                        *
 ********************************************************************/
/**
 * \brief This method is called when the application is closed.
 *
 * It stores the application state for the next session by calling
 * writeSettings().
 *
 * \param event The close event.
 * \see writeSettings()
 */
void MainWindow::closeEvent ( QCloseEvent *event )
{
    // if working if true the application was just started and is loading the
    // old files
    if ( working_ )
    {
        event->ignore ();
    } else
    {
        writeSettings ();
        ui_.console->stopExec();
        event->accept ();
    }
}

/**
 * \brief Slot which is called, if the data in a loaded file has changed
 *        on disk.
 *
 * \param file The path to the file.
 */
void MainWindow::fileChanged ( QString file )
{
    fileSystemWatcher.removePath ( file );
    QMessageBox msgBox ( this );

    msgBox.setWindowTitle ( "Unicode Viewer" );
    msgBox.setText ( tr ( "The document has been modified." ) );
    msgBox.setInformativeText (
            tr ( "Do you want reload the file %1 ?" ).arg ( file ) );

    msgBox.setIcon ( QMessageBox::Question );
    msgBox.setStandardButtons ( QMessageBox::Yes | QMessageBox::No );
    msgBox.setDefaultButton ( QMessageBox::No );

    int ret = msgBox.exec ();
    switch ( ret )
    {
        case QMessageBox::Yes:
            reloadFiles ();
            break;
        case QMessageBox::No:
            break;
    }
    fileSystemWatcher.addPath ( file );
}
/**
 * \brief Loads the files that were loaded when the application was closed the
 *        last time.
 *
 * \see loadTable()
 * \see loadBlockTable()
 */
void MainWindow::loadOldFiles ()
{
    if ( !curFile_.isEmpty () && QFile::exists ( curFile_.last () ) )
    {
        loadTable ( curFile_.last (), false );
    }
    if ( !curBlock_.isEmpty () && QFile::exists ( curBlock_ ) )
    {
        loadBlockTable ( curBlock_, false );
    }
}

/**
 * \brief Overwritten show() method to call loadOldFiles() and initLua() once
 *        the eventLoop is ready.
 *
 * \see initLua()
 * \see loadOldFiles()
 */
void MainWindow::show ()
{
    working_ = true;
    QMainWindow::show ();
    loadOldFiles ();
    initLua ();
    working_ = false;

}

/**
 * \brief Updates the states of the actions in the menu and the toolbars.
 */
void MainWindow::updateStates ()
{
    ui_.toolBarFont->enableActions    ( tableModel_->hasVisibleTable () );
    ui_.tableToolBar->enableActions   ( tableModel_->hasVisibleTable () );
    ui_.actionUcTableSave->setEnabled ( tableModel_->hasVisibleTable () );
    ui_.toolBarUc->enableSaveAction   ( tableModel_->hasVisibleTable () );
    ui_.menuSearch->setEnabled        ( tableModel_->hasVisibleTable () );
    ui_.actionSearch->setEnabled      ( tableModel_->hasVisibleTable () );
    ui_.actionZoomIn->setEnabled      ( tableModel_->hasVisibleTable () );
    ui_.actionZoomOut->setEnabled     ( tableModel_->hasVisibleTable () );
    ui_.actionAddUcTable->setEnabled  ( tableModel_->hasVisibleTable () );
    ui_.sortModeCombo->setEnabled     ( tableModel_->hasVisibleTable () );
    ui_.sortButton->setEnabled        ( tableModel_->hasVisibleTable () );
    ui_.sortOrderButton->setEnabled   ( tableModel_->hasVisibleTable () );


    ui_.actionAddBlocksTable->setEnabled   ( tableModel_->hasBlocks () );
    ui_.tableToolBar->enableAddBlockAction ( tableModel_->hasBlocks ()
                                             || tableModel_->hasTable () );

}

/**
 * \brief Slot for menu sort blocks by name ascending.
 */
void MainWindow::on_actionBlocksNameAsc_triggered()
{
    emit sortBlocks(UCQBlockList::SORT_NAME_ASCENDING);
}

/**
 * \brief Slot for menu sort blocks by name descending.
 */
void MainWindow::on_actionBlocksNameDesc_triggered()
{
    emit sortBlocks(UCQBlockList::SORT_NAME_DESCENDING);
}

/**
 * \brief Slot for menu sort blocks by code ascending.
 */
void MainWindow::on_actionBlocksCodeAsc_triggered()
{
    emit sortBlocks(UCQBlockList::SORT_CODE_ASCENDING);
}

/**
 * \brief Slot for menu sort blocks by code descending.
 */
void MainWindow::on_actionBlocksCodeDesc_triggered()
{
    emit sortBlocks(UCQBlockList::SORT_CODE_DESCENDING);
}

/**
 * \brief Opens a file dialog, loads the chosen font file and
 *        selects it as the font for the table.
 *
 * From the Qt Docs: "Currently only TrueType fonts, TrueType font collections,
 *                    and OpenType fonts are supported."
 */
void MainWindow::on_actionLoadFont_triggered (){
    QString file = QFileDialog::getOpenFileName (
            this, tr ( "Load Font" ), QDir::homePath (),
                tr ( "Font (*.ttf *.ttc *.otf)" ));

    if ( !file.isEmpty () && QFile::exists ( file ) )
    {
        // font id in db
        int id = QFontDatabase::addApplicationFont ( file );
        if(id < 0){
            statusBar()->showMessage(tr("Sorry - Font could not be loaded."),
                                     5000);
        } else {
            QStringList fams = QFontDatabase::applicationFontFamilies ( id );
            if(!fams.isEmpty()){
                ui_.toolBarFont->selectFont(QFont(fams.first()));
            }
        }
//        ui_.fontList->refresh();
    }
}

/**
 * \brief Sort for the sort order selector.
 *
 * Toggles the sort order and the appearance of the button.
 */
void MainWindow::on_sortOrderButton_clicked(){
    ascending_ = !ascending_;
    ui_.sortOrderButton->setIcon(ascending_ ? *sortAscIcon_ : *sortDescIcon_);
    ui_.sortOrderButton->setToolTip(ascending_ ? tr("Ascending order") : tr("Descending order"));
}

/**
 * \brief Slot for the sort button, triggers a sort.
 */
void MainWindow::on_sortButton_clicked(){
    switch(ui_.sortModeCombo->currentIndex()){
        case 0:
            sortControl_->sort(SortController::CODEPOINT_SORT);
            break;
        case 1:
            sortControl_->sort(SortController::COLLATION_SORT);
            break;
        case 2:
            sortControl_->sort(SortController::COMMENT_SORT);
            break;
    }
    ascending_ ? sortControl_->sortOrder(SortController::ASCENDING)
               : sortControl_->sortOrder(SortController::DESCENDING);
}

